%  rcs.sl			-*- SLang -*-
%
%  This file provides an interface to RCS a la Emacs (sort of).
%
%  Written by Guido Gonzato  <guido@ibogeo.df.unibo.it>
%  Last updated: 25 November 1998
%
%  To use this facility, put this in your .jedrc:
%  () = evalfile("rcs.sl");
%
%  Commands available:
%  ^X^F (rcs_open_file())    -- open an RCS file
%  ^Xvv (check_in_and_out()) -- check in or out an RCS file


static variable file, dir, name, flags, status, exit_status;

static define num_elements (name, delim)
{
  variable nth = 0;
  while (NULL != extract_element (name, nth, delim))
    nth++;
  return nth;
}

define rcs_open_file ()		% ^X^F
{
  variable rcs_file, cmd, n;
  
  file = read_file_from_mini ("RCS open file:");
  rcs_file = file;

  % check whether the file exists; if not, try the RCS version

  status = file_status (file);
  if (1 == status) {
    () = find_file (file);
    return;
  }
  
  % file not found. Build a file name like "/home/guido/RCS/file.txt,v"
  
  n = num_elements(file, '/');
  name = extract_element (rcs_file, n - 1, '/');
  while (str_replace (rcs_file, name, ""))
    rcs_file = ();
  rcs_file = Sprintf ("%s%s%s%s", rcs_file, "RCS/", name, ",v", 4);
  
  % now check if this file exists
  
  status = file_status (rcs_file);
  if (0 == status)
    {
      error ("RCS file not found.");
    }
  else 
    {
      cmd = Sprintf ("co %s > /dev/null 2>&1", name, 1);
      exit_status = run_shell_cmd (cmd);
      if (0 != exit_status) 
	{
	  error (Sprintf("Error checking out %s!", name, 1));
	  return;
	}
      () = find_file (name);
      flush ("Note: file is write protected.");
    }

}

define rcs_check_in_and_out ()	% ^X-v-v
{
  variable cmd, tmp, ch;

  USER_BLOCK0
    {
       () = write_buffer (file);
      tmp = read_mini ("Enter a change comment:", "", "");
      cmd = Sprintf ("echo \"%s\" | ci %s > /dev/null 2>&1", tmp, name, 2);
      exit_status = run_shell_cmd (cmd);
      if (0 != exit_status) 
	{
	  error (Sprintf("Error checking in %s!", name, 1));
	  return;
	}      
      setbuf_info (getbuf_info () | 0x8);
      flush ("Note: file is write protected.");
    }

  % check if the current buffer is attached to an RCS file.
  
  (file, dir, name, flags) = getbuf_info();
  tmp = Sprintf("%sRCS/%s,v", dir, file, 2);
  status = file_status (tmp);
  
  % if it doesn't exist, then create the RCS, check in, and exit
  
  if (0 == status) 
    {
      status = file_status ("RCS"); % the subdir RCS must exist
      if (0 == status) 
	{
#iftrue
	   exit_status = mkdir ("RCS", 0777);
#else
	  exit_status = run_shell_cmd ("mkdir RCS");
#endif
	  if (0 != exit_status) 
	    {
	      error ("Error creating RCS directory!");
	      return;
	    }
	}
      X_USER_BLOCK0;
      return;
    }
  
  % the RCS file exists; if the buffer is read only, then check it out
  
  if (flags & (1 shl 3)) % read only
    {
      cmd = Sprintf ("co -l %s > /dev/null 2>&1", name, 1);
      exit_status = run_shell_cmd (cmd);
      if (0 != exit_status) 
	{
	  error (Sprintf("Error checking out %s!", name, 1));
	  return;
	}
      delbuf (whatbuf());
      () = find_file (name);
      tmp = Sprintf ("%s%s", dir, file, 2);
      flush (Sprintf ("Checking out %s...done.", tmp, 1));
    }
  else % check it in
    {
      X_USER_BLOCK0;
    }
  
}

% these are the two commands

setkey ("rcs_open_file", 	"^X^F");
setkey ("rcs_check_in_and_out",	"^Xvv");

% --- End of file rcs.sl ---
