# jed-init.rb

require "buffer"
require "jed-misc"
# require "jed-aliases"

unless $BATCH
  stdout = Object.new
  def stdout.write(str)
    insert str
  end
  $> = stdout
end

def execute_extended_command
  if $MINIBUFFER_ACTIVE
    evaluate_cmd
    return
  end
  
  str = ""
  i = 0
  loop do
    # internal function
    if internal_function?(str)
      call_internal_function str
      return
    end
    
    # ruby function
    if str != "" && respond_to?(str, TRUE)
      eval str
      return
    end
    
    # S-Lang function
    if defined_in_slang?(str)
      eval_slang_string str
      return
    end
    
    unless $EXECUTING_MACRO or $DEFINING_MACRO
      ungetkey(13) if i == 1
      ungetkey(32)
      i += 1
    end
    
    str = read_with_completion("M-x", "", str, ?F)
  end
end
undefine_key "^[x"
define_key :execute_extended_command, "^[x"

MODE_TABLE = Hash.new

add_hook :mode_hook do |ext|
  if mode = MODE_TABLE[ext]
    eval(mode)
  end
end

def add_mode_for_extension(mode, ext)
  unless mode.is_a?(String) and ext.is_a?(String)
    raise TypeError, "arguments must be String"
  end
  MODE_TABLE[ext] = mode
end

def delete_mode_for_extension(ext)
  unless ext.is_a?(String)
    raise TypeError, "argument must be String"
  end
  MODE_TABLE[ext] = nil
end

require "ruby-mode"
add_mode_for_extension("ruby_mode", "rb")

jedrc_file = ENV["HOME"] + "/" + ".jedrc.rb"
if File.exists? jedrc_file
  load jedrc_file
end

add_completion("eval_buffer");
