# ruby-mode.rb

$ruby_mode_name = "ruby"
$ruby_mode_version = "1.4"

$ruby_indent_level = 2

def ruby_show_version
  messagef "Ruby mode [version %s]", $ruby_mode_version
end

def ruby_looking_at_keyword(keyword)
  save_excursion do
    if looking_at(keyword)
      forward_char(keyword.length)
      return looking_at_words(" ", "\t", ";") || end_of_line?
    else
      return FALSE
    end
  end
end

def ruby_looking_at_keywords(*keywords)
  keywords.each do |keyword|
    return TRUE if ruby_looking_at_keyword(keyword)
  end
  return FALSE
end

def ruby_calculate_indent
  $CASE_SEARCH = TRUE
  
  indent = 0
  extra_indent = 0
  
  save_excursion do
    beginning_of_line
    skip_whitespace
    indent = current_column
    if ruby_looking_at_keywords("end", "else", "elsif", "rescue",
				"ensure", "when") || looking_at_words("}", "]")
      
      extra_indent -= $ruby_indent_level
    end
    return indent if prevline(1) == 0
    
    end_of_line
    par_level = 0
    n = 0
    while TRUE
      if end_of_line?
	while TRUE
	  beginning_of_line
	  if looking_at("#")
	    return indent if prevline(1) == 0
	    end_of_line
	  else
	    end_of_line
	    break
	  end
	end
      end
      
      backward_char
      ch = current_char
      case ch
      when ?)
	par_level -= 1
      when ?(
	par_level += 1
	if par_level == 1
	  return current_column + 1
	end
      when ?{, ?[
	n += $ruby_indent_level
      when ?}, ?]
	n -= $ruby_indent_level
      end 
      
      if ruby_looking_at_keyword(" do")
	n += $ruby_indent_level
      elsif ruby_looking_at_keyword("end")
	n -= $ruby_indent_level
      end
      
      if (beginning_of_line? && par_level == 0) || beginning_of_buffer?
	skip_whitespace
	indent = current_column
	break
      end
    end
    
    if looking_at("#")
      return current_column
    end
    
    if ruby_looking_at_keywords("class", "module", "def", "if", "else",
				"elsif", "unless", "case", "when", "while",
				"until", "for", "begin", "rescue", "ensure")
      n += $ruby_indent_level
    end
    if n > 0
      extra_indent += n
    end
  end
  
  return indent + extra_indent
end

def ruby_indent_to(n)
  step = current_column
  beginning_of_line
  skip_whitespace
  step -= current_column
  if current_column != n
    beginning_of_line
    trim_whitespace
    n -= 1
    whitespace(n)
  end
  if step > 0
    forward_char(step)
  end
end

def ruby_indent_line
  ruby_indent_to(ruby_calculate_indent)
end

def ruby_newline_and_indent
  step = current_column
  beginning_of_line
  skip_whitespace
  step -= current_column
  if looking_at_words("end", "else", "elsif", "rescue",
		      "ensure", "when", "}")
    ruby_indent_line();
  end
  forward_char(step)
  newline
  ruby_indent_line
end

def ruby_self_insert_command
  return nil if $LAST_CHAR > 125
  insert($LAST_CHAR.chr)
  step = current_column
  beginning_of_line
  skip_whitespace
  step -= current_column
  if looking_at_words("end", "else", "elsif", "rescue",
		      "ensure", "when", "}")
    ruby_indent_line
  end
  forward_char(step)
  $LAST_CHAR
end

def ruby_mode
  set_mode($ruby_mode_name, LANGUAGE_MODE)
  use_keymap($ruby_mode_name)
  use_syntax_table($ruby_mode_name)
  run_hooks(:ruby_mode_hook)
end

unless keymap?($ruby_mode_name)
  create_keymap($ruby_mode_name)
end
define_key(:ruby_show_version, "^Cv", $ruby_mode_name)
define_key(:ruby_indent_line, "^I", $ruby_mode_name)
define_key(:ruby_newline_and_indent, "^M", $ruby_mode_name)
define_key(:eval_buffer, "^C^C", $ruby_mode_name)
define_key(:eval_region, "^X^E", $ruby_mode_name)
define_key(:ruby_self_insert_command, "0", $ruby_mode_name)
define_key(:ruby_self_insert_command, "1", $ruby_mode_name)
define_key(:ruby_self_insert_command, "2", $ruby_mode_name)
define_key(:ruby_self_insert_command, "3", $ruby_mode_name)
define_key(:ruby_self_insert_command, "4", $ruby_mode_name)
define_key(:ruby_self_insert_command, "5", $ruby_mode_name)
define_key(:ruby_self_insert_command, "6", $ruby_mode_name)
define_key(:ruby_self_insert_command, "7", $ruby_mode_name)
define_key(:ruby_self_insert_command, "8", $ruby_mode_name)
define_key(:ruby_self_insert_command, "9", $ruby_mode_name)
define_key(:ruby_self_insert_command, "a", $ruby_mode_name)
define_key(:ruby_self_insert_command, "b", $ruby_mode_name)
define_key(:ruby_self_insert_command, "c", $ruby_mode_name)
define_key(:ruby_self_insert_command, "d", $ruby_mode_name)
define_key(:ruby_self_insert_command, "e", $ruby_mode_name)
define_key(:ruby_self_insert_command, "f", $ruby_mode_name)
define_key(:ruby_self_insert_command, "g", $ruby_mode_name)
define_key(:ruby_self_insert_command, "h", $ruby_mode_name)
define_key(:ruby_self_insert_command, "i", $ruby_mode_name)
define_key(:ruby_self_insert_command, "j", $ruby_mode_name)
define_key(:ruby_self_insert_command, "k", $ruby_mode_name)
define_key(:ruby_self_insert_command, "l", $ruby_mode_name)
define_key(:ruby_self_insert_command, "m", $ruby_mode_name)
define_key(:ruby_self_insert_command, "n", $ruby_mode_name)
define_key(:ruby_self_insert_command, "o", $ruby_mode_name)
define_key(:ruby_self_insert_command, "p", $ruby_mode_name)
define_key(:ruby_self_insert_command, "q", $ruby_mode_name)
define_key(:ruby_self_insert_command, "r", $ruby_mode_name)
define_key(:ruby_self_insert_command, "s", $ruby_mode_name)
define_key(:ruby_self_insert_command, "t", $ruby_mode_name)
define_key(:ruby_self_insert_command, "u", $ruby_mode_name)
define_key(:ruby_self_insert_command, "v", $ruby_mode_name)
define_key(:ruby_self_insert_command, "w", $ruby_mode_name)
define_key(:ruby_self_insert_command, "x", $ruby_mode_name)
define_key(:ruby_self_insert_command, "y", $ruby_mode_name)
define_key(:ruby_self_insert_command, "z", $ruby_mode_name)
define_key(:ruby_self_insert_command, "A", $ruby_mode_name)
define_key(:ruby_self_insert_command, "B", $ruby_mode_name)
define_key(:ruby_self_insert_command, "C", $ruby_mode_name)
define_key(:ruby_self_insert_command, "D", $ruby_mode_name)
define_key(:ruby_self_insert_command, "E", $ruby_mode_name)
define_key(:ruby_self_insert_command, "F", $ruby_mode_name)
define_key(:ruby_self_insert_command, "G", $ruby_mode_name)
define_key(:ruby_self_insert_command, "H", $ruby_mode_name)
define_key(:ruby_self_insert_command, "I", $ruby_mode_name)
define_key(:ruby_self_insert_command, "J", $ruby_mode_name)
define_key(:ruby_self_insert_command, "K", $ruby_mode_name)
define_key(:ruby_self_insert_command, "L", $ruby_mode_name)
define_key(:ruby_self_insert_command, "M", $ruby_mode_name)
define_key(:ruby_self_insert_command, "N", $ruby_mode_name)
define_key(:ruby_self_insert_command, "O", $ruby_mode_name)
define_key(:ruby_self_insert_command, "P", $ruby_mode_name)
define_key(:ruby_self_insert_command, "Q", $ruby_mode_name)
define_key(:ruby_self_insert_command, "R", $ruby_mode_name)
define_key(:ruby_self_insert_command, "S", $ruby_mode_name)
define_key(:ruby_self_insert_command, "T", $ruby_mode_name)
define_key(:ruby_self_insert_command, "U", $ruby_mode_name)
define_key(:ruby_self_insert_command, "V", $ruby_mode_name)
define_key(:ruby_self_insert_command, "W", $ruby_mode_name)
define_key(:ruby_self_insert_command, "X", $ruby_mode_name)
define_key(:ruby_self_insert_command, "Y", $ruby_mode_name)
define_key(:ruby_self_insert_command, "Z", $ruby_mode_name)
define_key(:ruby_self_insert_command, "_", $ruby_mode_name)
define_key(:ruby_self_insert_command, "{", $ruby_mode_name)
define_key(:ruby_self_insert_command, "}", $ruby_mode_name)
define_key(:ruby_self_insert_command, ";", $ruby_mode_name)

create_syntax_table $ruby_mode_name
define_syntax($ruby_mode_name, ?%, "#", "")
define_syntax($ruby_mode_name, ?(, "([{<", ")]}>")
define_syntax($ruby_mode_name, ?", ?")
define_syntax($ruby_mode_name, ?', ?')
define_syntax($ruby_mode_name, ?\\, ?\\)
define_syntax($ruby_mode_name, ?w, "$0-9a-zA-Z_")
define_syntax($ruby_mode_name, ?0, "-+0-9a-fA-F.xXL")
define_syntax($ruby_mode_name, ?,, ",;.?:")
define_syntax($ruby_mode_name, ?+, "%-+/&*=<>|!~^")

if defined?(HAS_DFA_SYNTAX)
  enable_highlight_cache("ruby.dfa", $ruby_mode_name)
  define_highlight_rule("#.*$", "comment", $ruby_mode_name)
  define_highlight_rule("([\\$%&@\\*]|\\$#)[A-Za-z_0-9]+", "normal",
			$ruby_mode_name)
  define_highlight_rule("\\$([_\\./,\"\\\\#\\*\\?\\]\\[;!@:\\$<>\\(\\)" +
			"%=\\-~\\^\\|&`'\\+]|\\^[A-Z])", "normal",
			$ruby_mode_name)
  define_highlight_rule("[A-Za-z_][A-Za-z_0-9]*", "Knormal",
			$ruby_mode_name)
  define_highlight_rule("[0-9]+(\\.[0-9]+)?([Ee][\\+\\-]?[0-9]*)?",
			"number", $ruby_mode_name)
  define_highlight_rule("0[xX][0-9A-Fa-f]*", "number", $ruby_mode_name)
  define_highlight_rule("[\\(\\[\\{\\<\\>\\}\\]\\),;\\.\\?:]",
			"delimiter", $ruby_mode_name)
  define_highlight_rule("[%\\-\\+/&\\*=<>\\|!~\\^]", "operator",
			$ruby_mode_name)
  define_highlight_rule("-[A-Za-z]", "keyword0", $ruby_mode_name)
  define_highlight_rule("'[^']*'", "string", $ruby_mode_name)
  define_highlight_rule("'[^']*$", "string", $ruby_mode_name)
  define_highlight_rule("\"([^\"\\\\]|\\\\.)*\"", "string", $ruby_mode_name)
  define_highlight_rule("\"([^\"\\\\]|\\\\.)*\\\\?$", "string", $ruby_mode_name)
  define_highlight_rule("m?/([^/\\\\]|\\\\.)*/[gio]*", "string", $ruby_mode_name)
  define_highlight_rule("m/([^/\\\\]|\\\\.)*\\\\?$", "string", $ruby_mode_name)
  define_highlight_rule("s/([^/\\\\]|\\\\.)*(/([^/\\\\]|\\\\.)*)?/[geio]*",
			"string", $ruby_mode_name)
  define_highlight_rule("s/([^/\\\\]|\\\\.)*(/([^/\\\\]|\\\\.)*)?\\\\?$",
			"string", $ruby_mode_name)
  define_highlight_rule("(tr|y)/([^/\\\\]|\\\\.)*(/([^/\\\\]|\\\\.)*)?/[cds]*",
			"string", $ruby_mode_name)
  define_highlight_rule("(tr|y)/([^/\\\\]|\\\\.)*(/([^/\\\\]|\\\\.)*)?\\\\?$",
			"string", $ruby_mode_name)
  define_highlight_rule(".", "normal", $ruby_mode_name)
  build_highlight_table($ruby_mode_name)
end

define_keywords($ruby_mode_name, 0,
		"end",
		"class", "module",
		"def", "undef",
		"return",
		"if", "unless", "else", "elsif", "then",
		"not", "and", "or",
		"raise", "fail",
		"while", "loop",
		"for", "in",
		"case", "when",
		"self", "super", "nil",
		"alias",
		"until",
		"begin", "ensure", "rescue",
		"do", "yield",
		"break", "next", "continue", "retry", "redo",
		"include",
		"load", "require", "autoload")
define_keywords($ruby_mode_name, 1,
		"TRUE", "FALSE")

$ruby_shell_mode_name = "ruby shell"
$ruby_shell_buffer = "*ruby shell*"
$ruby_ps1 = "ruby> "
$ruby_ps2 = "ruby| "

Buffer.add_attr :ruby_prompt, $ruby_ps1
Buffer.add_attr :ruby_indent, 0
Buffer.add_attr :ruby_buff, ""

def ruby_get_current_line_cmd
  save_excursion do
    beginning_of_line
    push_mark
    end_of_line
    str = buffer_substring
    rexp = /^(#{Regexp.quote($ruby_ps1.strip)}|#{Regexp.quote($ruby_ps2.strip)})/
    if str =~ rexp
      str = str[$&.length .. -1]
    end
    str.strip
  end
end

def ruby_check_indent(line)
  if line =~ /^\s*(class|module|def|if|case|while|for|begin)\b[^_]/o
    $curbuf.ruby_indent = $curbuf.ruby_indent + 1
  end
  if line =~ /^\s*end\b[^_]/o
    $curbuf.ruby_indent = $curbuf.ruby_indent - 1
  end
  if line =~ /(\{|do)\s*(\|.*\|)?\s*$/o
    $curbuf.ruby_indent = $curbuf.ruby_indent + 1
  end
  if line =~ /^\s*\}/o
    $curbuf.ruby_indent = $curbuf.ruby_indent - 1
  end
end

def ruby_eval_and_newline
  end_of_line
  line = ruby_get_current_line_cmd
  if line == "exit"
    $curbuf.modified = FALSE
    $curbuf.delete
    return
  end
  unless end_of_buffer?
    end_of_buffer
    newline
    insert $curbuf.ruby_prompt, line
    return
  end
  line << "\n"
  newline
  $curbuf.ruby_buff << line
  if line =~ /,\s*$/o
    $curbuf.ruby_prompt = $ruby_ps2
    insert $curbuf.ruby_prompt
    return
  end
  ruby_check_indent(line)
  if $curbuf.ruby_indent > 0
    $curbuf.ruby_prompt = $ruby_ps2
    insert $curbuf.ruby_prompt
    return
  end
  begin
    p eval($curbuf.ruby_buff, TOPLEVEL_BINDING)
  rescue
    $! = "exception raised" unless $!
    print $!.type, ": ", $!, "\n"
  end
  $curbuf.ruby_buff = ""
  $curbuf.ruby_prompt = $ruby_ps1
  insert $curbuf.ruby_prompt
end

def ruby_beginning_of_line
  end_of_line
  unless end_of_buffer?
    beginning_of_line
    return
  end
   if current_column > $curbuf.ruby_prompt.length
     goto_column($curbuf.ruby_prompt.length + 1)
   else
     beginning_of_line
   end
end

def ruby_shell
  if buf = Buffer.find($ruby_shell_buffer)
      end_of_buffer
  else
    buf = Buffer.new($ruby_shell_buffer)
   end
   buf.show
  set_mode($ruby_shell_mode_name, LANGUAGE_MODE)
  use_keymap($ruby_shell_mode_name)
  use_syntax_table($ruby_mode_name)
  run_hooks(:ruby_shell_mode_hook)
  insert $curbuf.ruby_prompt
end

unless keymap?($ruby_shell_mode_name)
  create_keymap($ruby_shell_mode_name)
end
define_key(:ruby_eval_and_newline, "^M", $ruby_shell_mode_name)
define_key(:ruby_eval_and_newline, "^J", $ruby_shell_mode_name)
define_key(:ruby_beginning_of_line, "^A", $ruby_shell_mode_name)
