/* rbuffer.c
   Copyright (C) 1997 MAEDA Shugo

   class Buffer
   module BufferFunctions */

#include "config.h"
#include "jed-feat.h"
#include "slang.h"
#include "buffer.h"
#include "cmds.h"
#include "ledit.h"
#include "jruby.h"

VALUE cMark;
#ifdef JED_HAS_LINE_MARKS
VALUE cLineMark;
#endif /* JED_HAS_LINE_MARKS */
VALUE cBuffer;
VALUE mBufFunc;
VALUE rb_buf_list;
static ID id_data, id_init;

static void no_mark()
{
    rb_raise(rb_eRuntimeError, "can't find mark");
}

#define SETFLAG(flags, flag, on) \
flags = on ? (flags | flag) : (flags & ~flag)

#define GetMarkData(obj, mmt) {\
    Data_Get_Struct(obj, SLang_MMT_Type, mmt);\
    if (mmt == NULL) no_mark();\
}

static void free_mark(SLang_MMT_Type *mark)
{
    SLang_free_mmt(mark);
}

extern SLang_MMT_Type *jed_make_user_object_mark();

static VALUE mark_s_new(VALUE class)
{
    SLang_MMT_Type *mmt;
    VALUE obj;

    if ((mmt = jed_make_user_object_mark()) == NULL)
	rb_raise(rb_eRuntimeError, "failed to create mark");
    obj = Data_Wrap_Struct(class, 0, free_mark, mmt);
    
    return obj;    
}

#ifdef JED_HAS_LINE_MARKS
static VALUE linemark_s_new(VALUE class, VALUE color)
{
    SLang_MMT_Type *mmt;
    User_Mark_Type *um;
    VALUE obj;

    if ((mmt = jed_make_user_object_mark()) == NULL)
	rb_raise(rb_eRuntimeError, "failed to create mark");
    um = (User_Mark_Type *) SLang_object_from_mmt(mmt);
    um->m.flags |= JED_LINE_MARK | (NUM2INT(color) & MARK_COLOR_MASK);
    obj = Data_Wrap_Struct(class, 0, free_mark, mmt);

    return obj;    
}
#endif /* JED_HAS_LINE_MARKS */

static VALUE mark_go(VALUE obj)
{
    SLang_MMT_Type *mmt;
    User_Mark_Type *um;
    int ret = -1;

    GetMarkData(obj, mmt);
    um = (User_Mark_Type *) SLang_object_from_mmt(mmt);
    if (mark_valid (&um->m)) {
	if (CBuf != um->b) rb_raise(rb_eRuntimeError, "mark not in buffer");
	else
	    ret = goto_mark(&um->m);
    }
    return INT2FIX(ret);
}

static VALUE mark_move(VALUE obj)
{
    SLang_MMT_Type *mmt;
    User_Mark_Type *um;
    Mark *m;

    GetMarkData(obj, mmt);
    um = (User_Mark_Type *) SLang_object_from_mmt(mmt);
    m = &um->m;
    
    if (!mark_valid (m))
	rb_raise(rb_eRuntimeError, "invalid mark");
    if (CBuf != um->b)
	rb_raise(rb_eRuntimeError, "mark not in buffer");
   
    m->line = CLine;
    m->point = Point;
    m->n = LineNum + CBuf->nup;
    return Qtrue;
}

static VALUE mark_eq(VALUE obj, VALUE other)
{
    SLang_MMT_Type *mmt1, *mmt2;
    User_Mark_Type *um1,*um2;
    Mark *m1, *m2;

    if (!rb_obj_is_kind_of(other, cMark))
	return Qfalse;
    GetMarkData(obj, mmt1);
    GetMarkData(other, mmt2);
    um1 = (User_Mark_Type *) SLang_object_from_mmt(mmt1);
    um2 = (User_Mark_Type *) SLang_object_from_mmt(mmt2);
    m1 = &um1->m;
    m2 = &um2->m;
    if (um1->b != um2->b)
	return Qfalse;
    else if (m1->n != m2->n)
	return Qfalse;
    else if (m1->point != m2->point)
	return Qfalse;
    else
	return Qtrue;
}

static VALUE mark_cmp(VALUE obj, VALUE other)
{
    SLang_MMT_Type *mmt1, *mmt2;
    User_Mark_Type *um1,*um2;
    Mark *m1, *m2;

    if (!rb_obj_is_kind_of(other, cMark)) {
	rb_raise(rb_eTypeError, "<=> requires Mark (%s given)",
		 rb_class2name(CLASS_OF(other)));
    }
    GetMarkData(obj, mmt1);
    GetMarkData(other, mmt2);
    um1 = (User_Mark_Type *) SLang_object_from_mmt(mmt1);
    um2 = (User_Mark_Type *) SLang_object_from_mmt(mmt2);
    if (um1->b != um2->b) {
	rb_raise(rb_eArgError, "mark is not in the same buffer");
    }
    m1 = &um1->m;
    m2 = &um2->m;
    if (m1->n > m2->n)
	return INT2FIX(1);
    else if (m1->n < m2->n)
	return INT2FIX(-1);
    else if (m1->point > m2->point)
	return INT2FIX(1);
    else if (m1->point < m2->point)
	return INT2FIX(-1);
    else
	return INT2FIX(0);
}

static void no_buffer()
{
    rb_raise(rb_eRuntimeError, "can't find buffer");
}

static Buffer *get_buffer_data(obj)
{
    Buffer *buf;
    VALUE data;

    data = rb_ivar_get(obj, id_data);
    Data_Get_Struct(data, Buffer, buf);
    if (buf == NULL) no_buffer();
    return buf;
}

static void free_buffer(Buffer *buf)
{
    buf = NULL;
}

static void buffer_mark(Buffer *buf)
{
    rb_gc_mark(buf->obj);
}    

VALUE prep_buffer(VALUE class, Buffer *buf, VALUE name)
{
    VALUE obj, data;

    if (buf == NULL) {
	rb_raise(rb_eRuntimeError, "failed to create buffer");
    }

    obj = rb_obj_alloc(class);
    data = Data_Wrap_Struct(class, buffer_mark, free_buffer, buf);

    rb_ivar_set(obj, id_data, data);
    buf->obj = obj;
    rb_funcall(buf->obj, id_init, 1, name);
    rb_hash_aset(rb_buf_list, name, obj);

    return obj;    
}

static VALUE buffer_s_new(VALUE class, VALUE name)
{
    Buffer *buf;

    Check_Type(name, T_STRING);
    buf = make_buffer(RSTRING(name)->ptr, CBuf->dir, NULL);
    switch_to_buffer(buf);
    prep_buffer(class, buf, rb_str_new2(buf->name));
    return buf->obj;
}

static VALUE buffer_s_find(VALUE obj, VALUE name)
{
    Buffer *buf;
    
    Check_Type(name, T_STRING);
    buf = find_buffer(RSTRING(name)->ptr);
    if (buf == NULL)
      return Qnil;
    else
      return buf->obj;
}

static VALUE buffer_s_exists(VALUE class, VALUE name)
{
    Check_Type(name, T_STRING);
    return RBOOL(bufferp(RSTRING(name)->ptr));
}

static VALUE buffer_s_each(VALUE obj)
{
    Buffer *b;

    b = CBuf;
    do {
	rb_yield(b->obj);
	b = b->next;
    } while (b != CBuf);
}

static VALUE buffer_delete(VALUE obj)
{
    Buffer *buf;
    ID id_delete;
    
    id_delete = rb_intern("delete");
    buf = get_buffer_data(obj);
    /* Maybe buf is already NULL */
    if (buf && buf != MiniBuffer) {
	kill_buffer_cmd(buf->name);
    }
    rb_funcall(rb_buf_list, id_delete, 1, rb_str_new2(buf->name));
    return Qnil;
}

static VALUE buffer_get_name(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return rb_str_new2(buf->name);
}

static VALUE buffer_set_name(VALUE obj, VALUE name)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    Check_Type(name, T_STRING);
    safe_strcpy(buf->name, RSTRING(name)->ptr, sizeof(buf->name));

    return Qnil;
}

static VALUE buffer_get_file(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return rb_str_new2(buf->file);
}

static VALUE buffer_set_file(VALUE obj, VALUE file)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    Check_Type(file, T_STRING);
    safe_strcpy(buf->file, RSTRING(file)->ptr, sizeof(buf->file));

    return Qnil;
}

static VALUE buffer_get_dir(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return rb_str_new2(buf->dir);
}

static VALUE buffer_set_dir(VALUE obj, VALUE dir)
{
    Buffer *buf;
    extern char *expand_filename(char *);
    
    buf = get_buffer_data(obj);
    Check_Type(dir, T_STRING);
    safe_strcpy(buf->dir, RSTRING(dir)->ptr, sizeof (buf->dir));
    fixup_dir(buf->dir);
    strcpy(buf->dir, expand_filename(buf->dir));

    return Qnil;
}

static VALUE buffer_get_modified(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & BUFFER_MODIFIED);
}

static VALUE buffer_set_modified(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, BUFFER_MODIFIED, flag);
    return Qnil;
}

static VALUE buffer_get_auto_save(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & AUTO_SAVE_BUFFER);
}

static VALUE buffer_set_auto_save(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, AUTO_SAVE_BUFFER, flag);
    return Qnil;
}

static VALUE buffer_get_file_modified(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & FILE_MODIFIED);
}

static VALUE buffer_set_file_modified(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, FILE_MODIFIED, flag);
    return Qnil;
}

static VALUE buffer_get_read_only(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & READ_ONLY);
}

static VALUE buffer_set_read_only(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, READ_ONLY, flag);
    return Qnil;
}

static VALUE buffer_get_over_write_mode(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & OVERWRITE_MODE);
}

static VALUE buffer_set_over_write_mode(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, OVERWRITE_MODE, flag);
    return Qnil;
}

static VALUE buffer_get_undo_enabled(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & UNDO_ENABLED);
}

static VALUE buffer_set_undo_enabled(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, UNDO_ENABLED, flag);
    return Qnil;
}

static VALUE buffer_get_buried(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & BURIED_BUFFER);
}

static VALUE buffer_set_buried(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, BURIED_BUFFER, flag);
    return Qnil;
}

static VALUE buffer_get_auto_save_just_save(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & AUTO_SAVE_JUST_SAVE);
}

static VALUE buffer_set_auto_save_just_save(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, AUTO_SAVE_JUST_SAVE, flag);
    return Qnil;
}

static VALUE buffer_get_no_backup(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & NO_BACKUP_FLAG);
}

static VALUE buffer_set_no_backup(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, NO_BACKUP_FLAG, flag);
    return Qnil;
}

static VALUE buffer_get_binary(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & BINARY_FILE);
}

static VALUE buffer_set_binary(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, BINARY_FILE, flag);
    return Qnil;
}

static VALUE buffer_get_add_cr_on_write(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & ADD_CR_ON_WRITE_FLAG);
}

static VALUE buffer_set_add_cr_on_write(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, ADD_CR_ON_WRITE_FLAG, flag);
    return Qnil;
}

static VALUE buffer_get_abbrev_mode(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return RBOOL(buf->flags & ABBREV_MODE);
}

static VALUE buffer_set_abbrev_mode(VALUE obj, VALUE flag)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    SETFLAG(buf->flags, ABBREV_MODE, flag);
    return Qnil;
}

static VALUE buffer_get_tab(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return INT2FIX(buf->local_vars.tab);
}

static VALUE buffer_set_tab(VALUE obj, VALUE n)
{
    Buffer *buf;

    Check_Type(n, T_FIXNUM);
    buf = get_buffer_data(obj);
    buf->local_vars.tab = FIX2INT(n);
    if (buf == CBuf)
      Buffer_Local.tab = buf->local_vars.tab;
    return Qnil;
}

static VALUE buffer_get_selective_display(VALUE obj)
{
    Buffer *buf;
    
    buf = get_buffer_data(obj);
    return INT2FIX(buf->local_vars.sd);
}

static VALUE buffer_set_selective_display(VALUE obj, VALUE n)
{
    Buffer *buf;

    Check_Type(n, T_FIXNUM);
    buf = get_buffer_data(obj);
    buf->local_vars.sd = FIX2INT(n);
    if (buf == CBuf)
      Buffer_Local.sd = buf->local_vars.sd;
    return Qnil;
}

static VALUE buffer_activate(VALUE obj)
{
    Buffer *buf;

    buf = get_buffer_data(obj);
    switch_to_buffer(buf);
    return Qnil;
}

static VALUE buffer_show(VALUE obj)
{
    Buffer *buf;

    buf = get_buffer_data(obj);
    switch_to_buffer_cmd(buf->name);
    return Qnil;
}

static VALUE buffer_visible_p(VALUE obj)
{
    Buffer *buf;

    buf = get_buffer_data(obj);
    return RBOOL(buffer_visible(buf));
}

static VALUE buffunc_what_buffer(VALUE obj)
{
    return rb_str_new2(what_buffer());
}

static VALUE buffunc_switch_to_buffer(VALUE obj, VALUE name)
{
    Check_Type(name, T_STRING);
    switch_to_buffer_cmd(RSTRING(name)->ptr);
    return Qnil;
}

static VALUE buffunc_pop2buf(VALUE obj, VALUE name)
{
    Check_Type(name, T_STRING);
    return rb_str_new2(pop_to_buffer(RSTRING(name)->ptr));
}

static VALUE buffunc_set_buffer(VALUE obj, VALUE name)
{
    Check_Type(name, T_STRING);
    return RBOOL(set_buffer(RSTRING(name)->ptr));
}

static VALUE buffunc_erase_buffer(VALUE obj)
{
    return RBOOL(erase_buffer());
}

static VALUE buffunc_get_line(VALUE obj)
{
    return rb_str_new2(CLine->data);
}

static VALUE buffunc_begin_of_buffer(VALUE obj)
{
    return RBOOL(bob());
}

static VALUE buffunc_end_of_buffer(VALUE obj)
{
    return RBOOL(eob());
}

static VALUE buffunc_begin_of_line(VALUE obj)
{
    return RBOOL(bol());
}

static VALUE buffunc_end_of_line(VALUE obj)
{
    return RBOOL(eol());
}

static VALUE buffunc_begin_of_buffer_p(VALUE obj)
{
    return RBOOL(bobp());
}

static VALUE buffunc_end_of_buffer_p(VALUE obj)
{
    return RBOOL(eobp());
}

static VALUE buffunc_begin_of_line_p(VALUE obj)
{
    return RBOOL(bolp());
}

static VALUE buffunc_end_of_line_p(VALUE obj)
{
    return RBOOL(eolp());
}

static VALUE buffunc_prevline(VALUE obj, VALUE num)
{
    int n;

    Check_Type(num, T_FIXNUM);
    n = FIX2INT(num);
    return INT2FIX(prevline(&n));
}

static VALUE buffunc_nextline(VALUE obj, VALUE num)
{
    int n;

    Check_Type(num, T_FIXNUM);
    n = FIX2INT(num);
    return INT2FIX(nextline(&n));
}

static VALUE buffunc_forward_char(int argc, VALUE *argv,
				  VALUE obj)
{
    VALUE num;
    int n;
    
    if (rb_scan_args(argc, argv, "01", &num) == 0) {
	n = 1;
    } else {
	Check_Type(num, T_FIXNUM);
	n = FIX2INT(num);
    }
    return INT2FIX(forwchars(&n));
}

static VALUE buffunc_backward_char(int argc, VALUE *argv,
				   VALUE obj)
{
    VALUE num;
    int n;

    if (rb_scan_args(argc, argv, "01", &num) == 0) {
	n = 1;
    } else {
	Check_Type(num, T_FIXNUM);
	n = FIX2INT(num);
    }
    return INT2FIX(backwchars(&n));
}

static VALUE buffunc_insert(int argc, VALUE *argv,
			    VALUE obj)
{
    VALUE str = rb_str_new2("");
    VALUE str2;
    int i;
    
    for (i = 0; i < argc; i++) {
	if (TYPE(argv[i]) == T_STRING)
	    str2 = argv[i];
	else
	    str2 = rb_obj_as_string(argv[i]);
	str = rb_str_plus(str, str2);
    }
    insert_string(RSTRING(str)->ptr);

    return str;
}

static VALUE buffunc_insert_file(VALUE obj, VALUE file)
{
    Check_Type(file, T_STRING);
    return INT2FIX(insert_file(RSTRING(file)->ptr));
}

static VALUE buffunc_insert_buffer(VALUE obj, VALUE buffer)
{
    if (rb_obj_is_kind_of(buffer, cBuffer)) {
	Buffer *buf;

	buf = get_buffer_data(buffer);
	insert_buffer(buf);
    } else {
	Check_Type(buffer, T_STRING);
	insert_buffer_name(RSTRING(buffer)->ptr);
    }
    return Qnil;
}

static VALUE buffunc_yank(VALUE obj)
{
    return RBOOL(yank());
}

static VALUE buffunc_delete_region(VALUE obj)
{
    return RBOOL(delete_region());
}

static VALUE buffunc_kill_region(VALUE obj)
{
    return RBOOL(kill_region());
}

static VALUE buffunc_delete_char(VALUE obj)
{
    return RBOOL(delete_char_cmd());
}

static VALUE buffunc_delete_backward_char(VALUE obj)
{
    return RBOOL(backward_delete_char_cmd());
}

static VALUE buffunc_indent_line(VALUE obj)
{
    return RBOOL(indent_line());
}

static VALUE buffunc_newline(VALUE obj)
{
    return RBOOL(newline_cmd());
}

static VALUE buffunc_newline_and_indent(VALUE obj)
{
    return RBOOL(newline_and_indent());
}

static VALUE buffunc_insert_char(VALUE obj)
{
    return RBOOL(ins_char_cmd());
}

static VALUE buffunc_quoted_insert(VALUE obj)
{
    return RBOOL(quoted_insert());
}

static VALUE buffunc_kill_line(VALUE obj)
{
    return RBOOL(kill_line());
}

static VALUE buffunc_push_spot(VALUE obj)
{
    return RBOOL(push_spot());
}

static VALUE buffunc_pop_spot(VALUE obj)
{
    return RBOOL(pop_spot());
}

static VALUE buffunc_push_mark(VALUE obj)
{
    return RBOOL(push_mark());
}

static VALUE buffunc_pop_mark(VALUE obj, VALUE num)
{
    Check_Type(num, T_FIXNUM);
    return RBOOL(pop_mark(FIX2INT(num)));
}

static VALUE buffunc_duplicate_mark(VALUE obj)
{
    return RBOOL(dup_mark());
}

static VALUE buffunc_exchange_point_and_mark(VALUE obj)
{
    return RBOOL(exchange_point_mark());
}

static VALUE buffunc_goto_column(VALUE obj, VALUE num)
{
    int n;
    
    Check_Type(num, T_FIXNUM);
    n = FIX2INT(num);
    goto_column(&n);
    return Qnil;
}

static VALUE buffunc_goto_line(VALUE obj, VALUE num)
{
    int n;
    
    Check_Type(num, T_FIXNUM);
    n = FIX2INT(num);
    goto_line(&n);
    return Qnil;
}

static VALUE buffunc_skip_chars_forward(VALUE obj, VALUE chars)
{
    Check_Type(chars, T_STRING);
    skip_chars(RSTRING(chars)->ptr);
    return Qnil;
}

static VALUE buffunc_skip_chars_backward(VALUE obj, VALUE chars)
{
    Check_Type(chars, T_STRING);
    bskip_chars(RSTRING(chars)->ptr);
    return Qnil;
}

static VALUE buffunc_skip_whitespace(VALUE obj)
{
    skip_whitespace();
    return Qnil;
}

static VALUE buffunc_skip_word_chars_forward(VALUE obj)
{
    skip_word_chars();
    return Qnil;
}

static VALUE buffunc_skip_word_chars_backward(VALUE obj)
{
    bskip_word_chars();
    return Qnil;
}

static VALUE buffunc_skip_non_word_chars_forward(VALUE obj)
{
    skip_non_word_chars();
    return Qnil;
}

static VALUE buffunc_skip_non_word_chars_backward(VALUE obj)
{
    bskip_non_word_chars();
    return Qnil;
}

static VALUE buffunc_bufsubstr(VALUE obj)
{
    extern char *make_buffer_substring(int *np);

    int n;
    char *buff;
    if (CBuf->marks == NULL)
	rb_raise(rb_eRuntimeError, "set mark first");
    buff = make_buffer_substring(&n);
    if (buff == NULL)
	return Qnil;
    else
	return rb_str_new2(buff);
}

static VALUE buffunc_what_column(VALUE obj)
{
    return INT2FIX(calculate_column());
}

static VALUE buffunc_what_line(VALUE obj)
{
    return INT2FIX(LineNum);
}

static VALUE buffunc_what_char(VALUE obj)
{
    return INT2FIX(what_char());
}

static VALUE buffunc_jwhat_char(VALUE obj)
{
    return INT2FIX(jwhat_char());
}

static VALUE buffunc_looking_at(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    return RBOOL(looking_at(RSTRING(str)->ptr));
}

static VALUE buffunc_narrow_to_region(VALUE obj)
{
    return RBOOL(narrow_to_region());
}

static VALUE buffunc_narrow(VALUE obj)
{
    return RBOOL(narrow_to_lines());
}

static VALUE buffunc_widen(VALUE obj)
{
    return RBOOL(widen());
}

static VALUE buffunc_widen_region(VALUE obj)
{
    return RBOOL(widen_region());
}

static VALUE buffunc_push_narrow(VALUE obj)
{
    jed_push_narrow();
    return Qnil;
}

static VALUE buffunc_pop_narrow(VALUE obj)
{
    jed_pop_narrow();
    return Qnil;
}

static VALUE buffunc_write_buffer(VALUE obj, VALUE filename)
{
    Check_Type(filename, T_STRING);
    return RBOOL(write_buffer_cmd(RSTRING(filename)->ptr));
}

static VALUE buffunc_write_region(VALUE obj, VALUE filename)
{
    Check_Type(filename, T_STRING);
    return RBOOL(write_region_cmd(RSTRING(filename)->ptr));
}

static VALUE buffunc_search_forward(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    return INT2FIX(search_forward(RSTRING(str)->ptr));
}

static VALUE buffunc_search_backward(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    return INT2FIX(search_backward(RSTRING(str)->ptr));
}

static VALUE buffunc_search_line_forward(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    return INT2FIX(forward_search_line(RSTRING(str)->ptr));
}

static VALUE buffunc_search_line_backward(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    return INT2FIX(backward_search_line(RSTRING(str)->ptr));
}

static VALUE buffunc_re_search_forward(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    return INT2FIX(re_search_forward(RSTRING(str)->ptr));
}

static VALUE buffunc_re_search_backward(VALUE obj, VALUE str)
{
    Check_Type(str, T_STRING);
    return INT2FIX(re_search_backward(RSTRING(str)->ptr));
}

static VALUE buffunc_trim_whitespace(VALUE obj)
{
    trim_whitespace();
    return Qnil;
}

static VALUE buffunc_whitespace(VALUE obj, VALUE num)
{
    int n;
    
    Check_Type(num, T_FIXNUM);
    n = FIX2INT(num);
    insert_whitespace(&n);
    return Qnil;
}

static VALUE buffunc_get_current_indent(VALUE obj)
{
    int n;
    get_current_indent(&n);
    return INT2FIX(n);
}

static VALUE buffunc_indent_to(VALUE obj, VALUE num)
{
    Check_Type(num, T_FIXNUM);
    indent_to(FIX2INT(num));
    return Qnil;
}

static VALUE buffunc_create_mark(VALUE obj)
{
    return mark_s_new(cMark);
}

#ifdef JED_HAS_LINE_MARKS
static VALUE buffunc_create_line_mark(VALUE obj, VALUE color)
{
    return linemark_s_new(cLineMark, color);
}
#endif /* JED_HAS_LINE_MARKS */

static VALUE buffunc_goto_mark(VALUE obj, VALUE mark)
{
    if (!rb_obj_is_kind_of(mark, cMark)) {
	rb_raise(rb_eTypeError, "wrong argument type %s (expected Mark)",
		  TYPE(mark));
    }
    mark_go(mark);

    return Qnil;
}

static VALUE buffunc_move_mark(VALUE obj, VALUE mark)
{
    if (!rb_obj_is_kind_of(mark, cMark)) {
	rb_raise(rb_eTypeError, "wrong argument type %s (expected Mark)",
		  TYPE(mark));
    }
    mark_move(mark);

    return Qnil;
}

static VALUE curbuf_getter()
{
    return CBuf->obj;
}

static void curbuf_setter(VALUE buffer)
{
    Buffer *buf;
    
    if (!rb_obj_is_kind_of(buffer, cBuffer)) {
	rb_raise(rb_eRuntimeError, "$curbuf must be instance of Buffer");
    }
    buf = get_buffer_data(buffer);
    switch_to_buffer(buf);
}

void Init_rb_buffer()
{
    id_data = rb_intern("data");
    id_init = rb_intern("initialize");

    cMark = rb_define_class("Mark", rb_cObject);
    rb_include_module(cMark, rb_mComparable);
    rb_define_singleton_method(cMark, "new",
			       mark_s_new, 0);
    rb_define_method(cMark, "go", mark_go, 0);
    rb_define_method(cMark, "move", mark_move, 0);
    rb_define_method(cMark, "==", mark_eq, 1);
    rb_define_method(cMark, "<=>", mark_cmp, 1);

#ifdef JED_HAS_LINE_MARKS
    cLineMark = rb_define_class("LineMark", cMark);
    rb_define_singleton_method(cLineMark, "new",
			       linemark_s_new, 1);
#endif /* JED_HAS_LINE_MARKS */

    cBuffer = rb_define_class("Buffer", rb_cObject);
    rb_define_singleton_method(cBuffer, "new",
			       buffer_s_new, 1);
    rb_define_singleton_method(cBuffer, "find", buffer_s_find, 1);
    rb_define_singleton_method(cBuffer, "exists?", buffer_s_exists, 1);
    rb_define_singleton_method(cBuffer, "each", buffer_s_each, 0);
    rb_define_method(cBuffer, "delete", buffer_delete, 0);
    rb_define_method(cBuffer, "name", buffer_get_name, 0);
    rb_define_method(cBuffer, "name=", buffer_set_name, 1);
    rb_define_method(cBuffer, "file", buffer_get_file, 0);
    rb_define_method(cBuffer, "file=", buffer_set_file, 1);
    rb_define_method(cBuffer, "dir", buffer_get_dir, 0);
    rb_define_method(cBuffer, "dir=", buffer_set_dir, 1);

    rb_define_method(cBuffer, "modified",
		     buffer_get_modified, 0);
    rb_define_method(cBuffer, "modified=",
		     buffer_set_modified, 1);
    rb_define_method(cBuffer, "auto_save",
		     buffer_get_auto_save, 0);
    rb_define_method(cBuffer, "auto_save=",
		     buffer_set_auto_save, 1);
    rb_define_method(cBuffer, "file_modified",
		     buffer_get_file_modified, 0);
    rb_define_method(cBuffer, "file_modified=",
		     buffer_set_file_modified, 1);
    rb_define_method(cBuffer, "read_only",
		     buffer_get_read_only, 0);
    rb_define_method(cBuffer, "read_only=",
		     buffer_set_read_only, 1);
    rb_define_method(cBuffer, "over_write_mode",
		     buffer_get_over_write_mode, 0);
    rb_define_method(cBuffer, "over_write_mode=",
		     buffer_set_over_write_mode, 1);
    rb_define_method(cBuffer, "undo_enabled",
		     buffer_get_undo_enabled, 0);
    rb_define_method(cBuffer, "undo_enabled=",
		     buffer_set_undo_enabled, 1);
    rb_define_method(cBuffer, "buried",
		     buffer_get_buried, 0);
    rb_define_method(cBuffer, "buried=",
		     buffer_set_buried, 1);
    rb_define_method(cBuffer, "force_save",
		     buffer_get_auto_save_just_save, 0);
    rb_define_method(cBuffer, "force_save=",
		     buffer_set_auto_save_just_save, 1);
    rb_define_method(cBuffer, "no_backup",
		     buffer_get_no_backup, 0);
    rb_define_method(cBuffer, "no_backup=",
		     buffer_set_no_backup, 1);
    rb_define_method(cBuffer, "binary",
		     buffer_get_binary, 0);
    rb_define_method(cBuffer, "binary=",
		     buffer_set_binary, 1);
    rb_define_method(cBuffer, "add_cr",
		     buffer_get_add_cr_on_write, 0);
    rb_define_method(cBuffer, "add_cr=",
		     buffer_set_add_cr_on_write, 1);
    rb_define_method(cBuffer, "abbrev_mode",
		     buffer_get_abbrev_mode, 0);
    rb_define_method(cBuffer, "abbrev_mode=",
		     buffer_set_abbrev_mode, 1);

    rb_define_method(cBuffer, "tab", buffer_get_tab, 0);
    rb_define_method(cBuffer, "tab=", buffer_set_tab, 1);
    rb_define_method(cBuffer, "selective_display",
		     buffer_get_selective_display, 0);
    rb_define_method(cBuffer, "selective_display=",
		     buffer_set_selective_display, 1);
    rb_define_method(cBuffer, "activate", buffer_activate, 0);
    rb_define_method(cBuffer, "show", buffer_show, 0);
    rb_define_method(cBuffer, "visible?", buffer_visible_p, 0);

    mBufFunc = rb_define_module("BufferFunctions");
    rb_define_module_function(mBufFunc, "current_buffer",
			      buffunc_what_buffer, 0);
    rb_define_module_function(mBufFunc, "switch_to_buffer",
			      buffunc_switch_to_buffer, 1);
    rb_define_module_function(mBufFunc, "pop_to_buffer",
			      buffunc_pop2buf, 1);
    rb_define_module_function(mBufFunc, "set_buffer",
			      buffunc_set_buffer, 1);
    rb_define_module_function(mBufFunc, "erase_buffer",
			      buffunc_erase_buffer, 0);
    rb_define_module_function(mBufFunc, "get_line",
			      buffunc_get_line, 0);
    rb_define_module_function(mBufFunc, "beginning_of_buffer",
			      buffunc_begin_of_buffer, 0);
    rb_define_module_function(mBufFunc, "end_of_buffer",
			      buffunc_end_of_buffer, 0);
    rb_define_module_function(mBufFunc, "beginning_of_line",
			      buffunc_begin_of_line, 0);
    rb_define_module_function(mBufFunc, "end_of_line",
			      buffunc_end_of_line, 0);
    rb_define_module_function(mBufFunc, "beginning_of_buffer?",
			      buffunc_begin_of_buffer_p, 0);
    rb_define_module_function(mBufFunc, "end_of_buffer?",
			      buffunc_end_of_buffer_p, 0);
    rb_define_module_function(mBufFunc, "beginning_of_line?",
			      buffunc_begin_of_line_p, 0);
    rb_define_module_function(mBufFunc, "end_of_line?",
			      buffunc_end_of_line_p, 0);
    rb_define_module_function(mBufFunc, "prevline",
			      buffunc_prevline, 1);
    rb_define_module_function(mBufFunc, "nextline",
			      buffunc_nextline, 1);
    rb_define_module_function(mBufFunc, "forward_char",
			      buffunc_forward_char, -1);
    rb_define_module_function(mBufFunc, "backward_char",
			      buffunc_backward_char, -1);
    rb_define_module_function(mBufFunc, "insert",
			      buffunc_insert, -1);
    rb_define_module_function(mBufFunc, "insert_file",
			      buffunc_insert_file, 1);
    rb_define_module_function(mBufFunc, "insert_buffer",
			      buffunc_insert_buffer, 1);
    rb_define_module_function(mBufFunc, "yank",
			      buffunc_yank, 0);
    rb_define_module_function(mBufFunc, "delete_region",
			      buffunc_delete_region, 0);
    rb_define_module_function(mBufFunc, "kill_region",
			      buffunc_kill_region, 0);
    rb_define_module_function(mBufFunc, "delete_char",
			      buffunc_delete_char, 0);
    rb_define_module_function(mBufFunc, "delete_backward_char",
			      buffunc_delete_backward_char, 0);
    rb_define_module_function(mBufFunc, "indent_line",
			      buffunc_indent_line, 0);
    rb_define_module_function(mBufFunc, "newline",
			      buffunc_newline, 0);
    rb_define_module_function(mBufFunc, "newline_and_indent",
			      buffunc_newline_and_indent, 0);
    rb_define_module_function(mBufFunc, "insert_char",
			      buffunc_insert_char, 0);
    rb_define_module_function(mBufFunc, "quoted_insert",
			      buffunc_quoted_insert, 0);
    rb_define_module_function(mBufFunc, "kill_line",
			      buffunc_kill_line, 0);
    rb_define_module_function(mBufFunc, "push_spot",
			      buffunc_push_spot, 0);
    rb_define_module_function(mBufFunc, "pop_spot",
			      buffunc_pop_spot, 0);
    rb_define_module_function(mBufFunc, "push_mark",
			      buffunc_push_mark, 0);
    rb_define_module_function(mBufFunc, "pop_mark",
			      buffunc_pop_mark, 1);
    rb_define_module_function(mBufFunc, "duplicate_mark",
			      buffunc_duplicate_mark, 0);
    rb_define_module_function(mBufFunc, "exchange_point_and_mark",
			      buffunc_exchange_point_and_mark, 0);
    rb_define_module_function(mBufFunc, "goto_column",
			      buffunc_goto_column, 1);
    rb_define_module_function(mBufFunc, "goto_line",
			      buffunc_goto_line, 1);
    rb_define_module_function(mBufFunc, "skip_chars_forward",
			      buffunc_skip_chars_forward, 1);
    rb_define_module_function(mBufFunc, "skip_chars_backward",
			      buffunc_skip_chars_backward, 1);
    rb_define_module_function(mBufFunc, "skip_whitespace",
			      buffunc_skip_whitespace, 0);
    rb_define_module_function(mBufFunc, "skip_word_chars_forward",
			      buffunc_skip_word_chars_forward, 0);
    rb_define_module_function(mBufFunc, "skip_word_chars_backward",
			      buffunc_skip_word_chars_backward, 0);
    rb_define_module_function(mBufFunc, "skip_non_word_chars_forward",
			      buffunc_skip_non_word_chars_forward, 0);
    rb_define_module_function(mBufFunc, "skip_non_word_chars_backward",
			      buffunc_skip_non_word_chars_backward, 0);
    rb_define_module_function(mBufFunc, "buffer_substring",
			      buffunc_bufsubstr, 0);
    rb_define_module_function(mBufFunc, "current_column",
			      buffunc_what_column, 0);
    rb_define_module_function(mBufFunc, "current_line",
			      buffunc_what_line, 0);
    rb_define_module_function(mBufFunc, "current_char",
			      buffunc_what_char, 0);
    rb_define_module_function(mBufFunc, "jcurrent_char",
			      buffunc_jwhat_char, 0);
    rb_define_module_function(mBufFunc, "looking_at",
			      buffunc_looking_at, 1);
    rb_define_module_function(mBufFunc, "narrow_to_region",
			      buffunc_narrow_to_region, 0);
    rb_define_module_function(mBufFunc, "narrow",
			      buffunc_narrow, 0);
    rb_define_module_function(mBufFunc, "widen",
			      buffunc_widen, 0);
    rb_define_module_function(mBufFunc, "widen_region",
			      buffunc_widen_region, 0);
    rb_define_module_function(mBufFunc, "push_narrow",
			      buffunc_push_narrow, 0);
    rb_define_module_function(mBufFunc, "pop_narrow",
			      buffunc_pop_narrow, 0);
    rb_define_module_function(mBufFunc, "write_buffer",
			      buffunc_write_buffer, 1);
    rb_define_module_function(mBufFunc, "write_region",
			      buffunc_write_region, 1);
    rb_define_module_function(mBufFunc, "search_forward",
			      buffunc_search_forward, 1);
    rb_define_module_function(mBufFunc, "search_backward",
			      buffunc_search_backward, 1);
    rb_define_module_function(mBufFunc, "search_line_forward",
			      buffunc_search_line_forward, 1);
    rb_define_module_function(mBufFunc, "search_line_backward",
			      buffunc_search_line_backward, 1);
    rb_define_module_function(mBufFunc, "re_search_forward",
			      buffunc_re_search_forward, 1);
    rb_define_module_function(mBufFunc, "re_search_backward",
			      buffunc_re_search_backward, 1);
    rb_define_module_function(mBufFunc, "trim_whitespace",
			      buffunc_trim_whitespace, 0);
    rb_define_module_function(mBufFunc, "whitespace",
			      buffunc_whitespace, 1);
    rb_define_module_function(mBufFunc, "get_current_indent",
			      buffunc_get_current_indent, 0);
    rb_define_module_function(mBufFunc, "indent_to",
			      buffunc_indent_to, 1);
    rb_define_module_function(mBufFunc, "create_mark",
			      buffunc_create_mark, 0);
#ifdef JED_HAS_LINE_MARKS
    rb_define_module_function(mBufFunc, "create_line_mark",
			      buffunc_create_line_mark, 1);
#endif /* JED_HAS_LINE_MARKS */
    rb_define_module_function(mBufFunc, "goto_mark",
			      buffunc_goto_mark, 1);
    rb_define_module_function(mBufFunc, "move_mark",
			      buffunc_move_mark, 1);
    rb_include_module(rb_cObject, mBufFunc);

    rb_buf_list = rb_hash_new();
    rb_global_variable(&rb_buf_list);
    rb_define_virtual_variable("$curbuf", curbuf_getter, curbuf_setter);
}
